// Digikey Sort By Price: Uses AJAX to fetch pages from digikey, sorts by price,
//                        and filters the table
// version 001, 2009-05-30
// Copyright (c) 2009, Alex Leone (acleone ~at~ gmail.com)
// Released under the GPL license
// http://www.gnu.org/copyleft/gpl.html
//
// --------------------------------------------------------------------
//
// This is a Greasemonkey user script.  To install it, you need
// Greasemonkey 0.3 or later: http://greasemonkey.mozdev.org/
// Then restart Firefox and revisit this script.
// Under Tools, there will be a new menu item to "Install User Script".
// Accept the default configuration and install.
//
// To uninstall, go to Tools/Manage User Scripts,
// select "Digikey Sort By Price", and click Uninstall.
//
// --------------------------------------------------------------------
//
// ==UserScript==
// @name          Digikey Sort By Price
// @namespace     http://students.washington.edu/acleone/
// @description   sorts by price after using AJAX to fetch all pages
// @include       http://search.digikey.com/scripts/DkSearch/dksus.dll*
// ==/UserScript==

/** tableData:
    tableData[x][0] - innerHTML of the row.
    tableData[x][1:] - textContent of the column, 1-indexed
*/  
var tableData;
var filteredTableData;  // copy of tableData except filtered
var filterFunctions;    // currently applied filters
var filterEnableChecks; // list of checkboxes that enable specific filters

var tableDisplay;

var tableHeadersHTML; // '<table><thead><tr>...</tr><tr>...</tr></thead><tbody>'
var priceCol;       // the price column index (1-indexed)
var infoDiv;        // a div just above the parts table which will display info
var filterInfo;     // div after the 'Update Filters' button to display info
var sortCol;        // which column to sort the parts table by (1 indexed)
var sortOrder;      // 'a' or 'd'
var colNames;       // textContent of column headers, 1-indexed

var FILTER_FUNCTIONS = {
    'inStock': function (row) {
        return parseFloat(row[priceCol - 2]) > 0 || row[priceCol - 2] === 'Available';
    },
	'numericPriceOnly': function (row) {
		return !isNaN(parseFloat(row[priceCol]));
	},
	'maxPriceEnable': function (row) {
		return ((parseFloat(row[priceCol]) * parseFloat(row[priceCol - 1])) <
		        document.getElementById('maxPrice').value);
	}
};

// called after all the pages have been put into the table (partTableTBody)
function allFetched() {
    infoDiv.textContent = 'Done fetching. Got ' + tableData.length + ' parts.';
	filterTableData();
	sortTable();
	redisplayPartsTable();
}

function sortByColAsc(row1, row2) {
    var data1 = row1[sortCol];
    var data2 = row2[sortCol];
    var n1 = parseFloat(data1, 10);
    var n2 = parseFloat(data2, 10);
    var r = n1 - n2;
    if (!isNaN(r)) { // two numbers
        return r;
    }
    if (isNaN(n1)) {
        if (isNaN(n2)) { // string comparison
            if (data1 > data2) {
                return 1;
            } else if (data1 < data2) { 
                return -1;
            }
            return 0;
        }
        return 1; // data2 is a number
    }
    return -1; // data1 is a number
}

function sortByColDsc(row2, row1) {
    var data1 = row1[sortCol];
    var data2 = row2[sortCol];
    var n1 = parseFloat(data1, 10);
    var n2 = parseFloat(data2, 10);
    var r = n1 - n2;
    if (!isNaN(r)) { // two numbers
        return r;
    }
    if (isNaN(n1)) {
        if (isNaN(n2)) { // string comparison
            if (data1 > data2) {
                return 1;
            } else if (data1 < data2) { 
                return -1;
            }
            return 0;
        }
        return -1; // data2 is a number
    }
    return 1; // data1 is a number
}

// sorts filteredTableData
function sortTable() {
    filteredTableData.sort((sortOrder === 'a')? sortByColAsc : sortByColDsc);
}

// updates filterFunctions with the functions currently enabled
function updateFilters(event) {
    filterFunctions = [];
    for (var i = 0; i < filterEnableChecks.length; i++) {
        var check = filterEnableChecks[i];
        if (check.checked) {
            filterFunctions.push(FILTER_FUNCTIONS[check.name]);
        }          
    }
    if (event) {
        event.preventDefault();
        filterTableData();
    	sortTable();
	    redisplayPartsTable();
    }
}

// updates filteredTableData by filtering on all the filterFunctions
function filterTableData() {
	filteredTableData = tableData;
	for (var i = 0; i < filterFunctions.length; i++) {
		filteredTableData = filteredTableData.filter(filterFunctions[i]);
	}
}

// replaces parts table by cloning all the nodes from filteredTableData
function redisplayPartsTable() {
	var newHTML = tableHeadersHTML;
	for (var i = 0; i < filteredTableData.length; i++) {
        newHTML += filteredTableData[i][0];
    }
	tableDisplay.innerHTML = newHTML + '</tbody></table>';
	filterInfo.textContent = 'Displaying ' + filteredTableData.length +
	        ' of ' + tableData.length + ' parts, sorted by \'' +
	        colNames[sortCol] +
	        ((sortOrder === 'a')? '\' ascending' : '\' descending');
	sortLinks = document.getElementById('sortRow').getElementsByTagName('a');
	for (i = 0; i < sortLinks.length; i++) {
	    sortLinks[i].addEventListener('click', sortBy, true);
	}
}

// onclick for the sort links
function sortBy(event) {
    if (event) {
        event.preventDefault();
    }
    var tSplit = this.id.split('_');
    sortCol = parseInt(tSplit[0], 10);
    sortOrder = tSplit[1];
    infoDiv.textContent = 'Sorting...';
    setTimeout(function () {
            sortTable();
            redisplayPartsTable();
            infoDiv.textContent = 'Done.';
    }, 0);
}

// we will search the page source for this to determine if we are on the
// list of parts page
function isPartsList(pagetext) {
    return pagetext.indexOf('help04.html') !== -1;
}

// adds to tableData
function addTableData(pagetext) {
	var tbodyStartI = pagetext.indexOf('<tbody>');
	if (tbodyStartI === -1) {
		console.log('tbody null');
		return false;
	}
	var tbodyStopI = pagetext.indexOf('</table>', tbodyStartI);
	var tabletext = (tbodyStopI === -1)? pagetext.substring(tbodyStartI) :
			                       pagetext.substring(tbodyStartI, tbodyStopI);
	var rows = tabletext.match(/<tr[\s\S]*?<\/tr>/gim);
	if (rows === null) {
		console.log('rows null');
		return false;
	}
	var div = document.createElement('div');
	for (var i = 0; i < rows.length; i++) {
		var row = rows[i];
		var rowData = [row];
		var colRE = /<td[\s\S]*?>([\s\S]*?)<\/td>/gim;
		var colMatch;
		var j = 1;
		while ((colMatch = colRE.exec(row)) !== null) {
			div.innerHTML = colMatch[1];
			rowData[j++] = div.textContent.replace(/,/g, '');
		}
		tableData.push(rowData);
	}
}

// called after a successful ajax fetch of the page
function parsePartsTable(xhr, pageNum) {
    var pagetext = xhr.responseText;
    if (!isPartsList(pagetext)) {
        return;
    }
    addTableData(pagetext);
}

// called after an error occured in our ajax fetch
function ajaxError(xhr) {
    console.log(xhr);
}

/** parses the first page: (pagetext is document.body.innerHTML)
    - removes the "page 1/n (next)" text
    - adds divs for info display
    - starts the page fetcher which will fetch the rest of the pages
*/
function parseFirstPage() {
    var contentDiv = document.getElementById('content');
	if (contentDiv === null) {
		return;
	}
	var ps = contentDiv.getElementsByTagName('p');
	if (ps.length === 0) {
		return;
	}
	var lastpAs = ps[ps.length - 1].getElementsByTagName('a');
	if (lastpAs.length === 0) {
		return;
	}
	if (lastpAs[0].href.indexOf('help04.html') === -1) {
		return;
	}
	
	var partsTable = contentDiv.getElementsByTagName('table')[0];
	
	var ptHead = partsTable.getElementsByTagName('thead')[0];
	var headers = ptHead.getElementsByTagName('tr')[0];
	var colNamesTHs = headers.getElementsByTagName('th');
	colNames = [''];
	for (var i = 0; i < colNamesTHs.length; i++) {
	    colNames[i + 1] = colNamesTHs[i].textContent;
    }
	tableHeadersHTML = '<table cellspacing="1" cellpadding="2" border="1">' +
					   '<thead><tr>' + headers.innerHTML + '</tr>' +
					   '<tr id="sortRow">';
	priceCol = headers.getElementsByTagName('th').length - 1;
	sortCol = priceCol;
	sortOrder = 'a';
	for (i = 0; i < priceCol; i++) {
		tableHeadersHTML += '<td><a href="#" id="' + (i + 1) + '_a" style="' +
				'padding: 0 0.1em; font-weight: bold; text-decoration: none;' +
				'float: left; color: rgb(0, 0, 255);">^</a><a href="#" ' +
				'id="' + (i + 1) + '_d" style="padding: 0pt 0.1em;' +
				'font-weight: bold; text-decoration: none; float: right;' +
				'color: rgb(0, 0, 255);">v</a></td>';
	}
	tableHeadersHTML += '<td>&nbsp;</td></tr></thead><tbody>';
	tableData = [];
	addTableData(partsTable.innerHTML);
	
	var pageMatch = partsTable.nextSibling.textContent.match(/Page (.*?)\/([^\s]*)/i);
	if (pageMatch !== null) {
		var pagenextForm = document.getElementsByName('pagenext')[0];
		var pagenextInputs = pagenextForm.getElementsByTagName('input');
		var nextNames = [];
		var nextValues = [];
		for (i = 0; i < pagenextInputs.length; i++) {
			var input = pagenextInputs[i];
			nextNames.push(input.name);
			nextValues.push(input.value);
		}
	}
	//---filterTable.html:---
	contentDiv.innerHTML = '<p id="infoDiv" style="font-size: 130%; font-weight: bold;">Loading...</p><div id="filterDiv"><form id="filtersForm" name="filtersForm" action="#"><fieldset><legend>Filter Table By:</legend><table border="0"><tr><td><input name="inStock" type="checkbox" checked="checked" /></td><td>In Stock Only (Quantity Available &gt; 0 or \'Available\')</td></tr><tr><td><input name="numericPriceOnly" type="checkbox" checked="checked" /></td><td>Only numeric prices (no \'Call\' or \'Calculate\')</td></tr><tr><td><input name="maxPriceEnable" type="checkbox" checked="checked" /></td><td>Maximum Price (Min Qty * Price): &nbsp;<input id="maxPrice" size="10" type="text" value="100.0" /></td></tr></table><input type="submit" value="Update Filters" /> &nbsp;<span id="filterInfo"></span></fieldset></form></div><div id="tableDisplay">Loading table...</div>';
	filterInfo = document.getElementById('filterInfo');
	var filtersForm = document.getElementById('filtersForm');
	filtersForm.addEventListener('submit', updateFilters, true);
    var filterInputs = filtersForm.getElementsByTagName('input');
    filterEnableChecks = [];
    for (i = 0; i < filterInputs.length; i++) {
        var inp = filterInputs[i];
        if (inp.type === 'checkbox' && FILTER_FUNCTIONS[inp.name]) {
            filterEnableChecks.push(inp);
        }
    }
    updateFilters();

	infoDiv = document.getElementById('infoDiv');
	tableDisplay = document.getElementById('tableDisplay');
    if (pageMatch === null) {
        allFetched();
        return;
    }
    var currentPage = parseInt(pageMatch[1].replace(/,/g, ''), 10);
    var outOf = parseInt(pageMatch[2].replace(/,/g, ''), 10);
    var requestList = [];
	var pageI = nextNames.indexOf('page');
    for (var pageNum = 1; pageNum <= outOf; pageNum++) {
        if (pageNum === currentPage) {
			continue;
		}
        nextValues[pageI] = pageNum;
        var postDataList = [];
        for (i = 0; i < nextNames.length; i++) {
            postDataList.push(nextNames[i] + '=' + nextValues[i]);
        }
        requestList.push({
            request: {
                method: 'POST',
                url: window.location.href,
                headers: {'Content-type': 'application/x-www-form-urlencoded'},
                data: encodeURI(postDataList.join('&'))
            },
            onload: parsePartsTable,
            onerror: ajaxError,
            params: pageNum
        });
    }   
    var pageFetcher = new PageFetcher(requestList, allFetched, infoDiv);
    pageFetcher.startFetching();
}

// requestList -- [
//         {
//  request: request object for GM_xmlhttpRequest (onload will be overwritten),
//  onload: callback/onload function (params: func(xhr, params)),
//  params: extra params passed to the callback
//         }, ... ]
//  statusDiv -- status messages will be displayed here
function PageFetcher(requestList, allDoneCallback, statusDiv) {
    this.MAX_CONCUR_FETCHES = 5;
    this.currentlyFetching = 0;
    this.toBeFetched = requestList;
    this.stopFetches = false;
    this.allDoneCallback = allDoneCallback;
    this.statusDiv = statusDiv;
    this.statusText = null;
    this.stopButton = null;
    this.totalFetches = 0;
    
    this.startFetching = function () {
        this.totalFetches = this.toBeFetched.length;
        this.statusDiv.innerHTML = '';
        this.statusText = document.createTextNode('Loading...');
        this.updateStatusText();
        this.statusDiv.appendChild(this.statusText);
        this.stopButton = document.createElement('button');
        this.stopButton.textContent = 'Stop';
        var self = this;
        this.stopButton.addEventListener('click', function () {
            if (self.stopFetches === false) {
                self.stopFetches = true;
                this.disabled = true;
            } else {
                this.textContent = 'Stop';
                self.stopFetches = false;
                self.generateFetches();
            }
        }, true);
        this.statusDiv.appendChild(this.stopButton);
        this.generateFetches();
    };
    
    this.updateStatusText = function () {
        this.statusText.textContent = 'Fetched ' + 
                (this.totalFetches - this.toBeFetched.length - this.currentlyFetching) +
                ' / ' + this.totalFetches + ' Pages...';
    };
    
    // defines the onload of the ajax request so that toFetch has the right
    // value when the onload function is called
    this.fakeCurry = function (self, toFetch) {
        toFetch.request.onload = function (xhr) {
            self.currentlyFetching--;
            self.updateStatusText();
            if (self.toBeFetched.length === 0 && self.currentlyFetching === 0) {
                toFetch.onload(xhr, toFetch.params);
                self.allDoneCallback();
            } else {
                self.generateFetches();
                setTimeout(toFetch.onload, 0, xhr, toFetch.params);
            }
        };
    };    
    
    // fetches more pages up to MAX_CONCUR_FETCHES
    this.generateFetches = function () {
        if (this.stopFetches) {
            if (this.currentlyFetching === 0) {
                this.stopButton.textContent = 'Continue';
                this.stopButton.disabled = false;
                filterTableData();
	            sortTable();
	            redisplayPartsTable();
            }
            return;
        }
        var numToFetch = Math.min(this.toBeFetched.length,
                             this.MAX_CONCUR_FETCHES - this.currentlyFetching);
        for (var i = 0; i < numToFetch; i++) {
            var toFetch = this.toBeFetched.shift();
            this.fakeCurry(this, toFetch);
            this.currentlyFetching++;
            GM_xmlhttpRequest(toFetch.request);
        }
    };
}

parseFirstPage();
